-- Lua optimizations
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strjoin = strjoin
local strlen = strlen
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tonumber = tonumber
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"
local FONTEND = FONT_COLOR_CODE_CLOSE
local NORMALFONT = NORMAL_FONT_COLOR_CODE
local YELLOWFONT = LIGHTYELLOW_FONT_COLOR_CODE
local GRAYFONT = GRAY_FONT_COLOR_CODE

local DELIMITER = "%"
local Indicators = {
	ICON_STAR = ICON_LIST[1] .. "16:16:0:0|t",
	ICON_CIRCLE = ICON_LIST[2] .. "16:16:0:0|t",
	ICON_DIAMOND = ICON_LIST[3] .. "16:16:0:0|t",
	ICON_TRIANGLE = ICON_LIST[4] .. "16:16:0:0|t",
	ICON_MOON = ICON_LIST[5] .. "16:16:0:0|t",
	ICON_SQUARE = ICON_LIST[6] .. "16:16:0:0|t",
	ICON_CROSS = ICON_LIST[7] .. "16:16:0:0|t",
	ICON_SKULL = ICON_LIST[8] .. "16:16:0:0|t",
	BLANK = "None",
	TEXT = "Text",
}

local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local QTipC = LibStub("LibQTipClick-1.1")
local tooltip, tooltipchild

local thisToon = strjoin(" - ", UnitName("player"), GetRealmName())
local selectedInstance -- for re-ordering the instance list

local db
local defaultDB = {
	DBVersion = 5,
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
	Indicators = {
		NormalIndicator = "ICON_STAR", -- indicator: ICON_*, BLANK, TEXT
		NormalText = "N",
		NormalColor = { 0, 1, 0, 1, }, -- green
		HeroicIndicator = "ICON_SKULL", -- indicator
		HeroicText = "H",
		HeroicColor = { 1, 0, 0, 1, }, -- red
	},
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
		NewInstanceShow = false,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	MinimapIcon = { },
}

-- general helper functions below

local function ColorCodeOpen(color)
	return format("|c%02x%02x%02x%02x", math.floor(color[4] * 255), math.floor(color[1] * 255), math.floor(color[2] * 255), math.floor(color[3] * 255))
end

local function ClassColorise(class, targetstring)
	local color = {
		RAID_CLASS_COLORS[class].r,
		RAID_CLASS_COLORS[class].g,
		RAID_CLASS_COLORS[class].b,
		1,
	}
	return strjoin("", ColorCodeOpen(color), targetstring, FONTEND)
end

local function TableLen(table)
	local i = 0
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end

local function OrderedInstances()
	local list = { }
	for instance, a in pairs(SavedInstances.db.Instances) do
		list[a.Order] = instance
	end
	return list
end

local function SetInstanceOrder(from, to)
	for k, v in pairs(SavedInstances.db.Instances) do
		if v.Order == from then
			v.Order = to
		elseif to < from and v.Order >= to and v.Order < from then
			v.Order = v.Order + 1
		elseif to > from and v.Order <= to and v.Order > from then
			v.Order = v.Order - 1
		end
	end
end

-- local addon functions below

local function DifficultyString(toon)
	local normal, heroic
	local prefs = SavedInstances.db.Indicators
	if toon.Normal and prefs.NormalIndicator ~= "BLANK" then
		if prefs.NormalIndicator == "TEXT" then
			normal = ColorCodeOpen(prefs.NormalColor) .. prefs.NormalText .. FONTEND
		elseif prefs.NormalIndicator ~= "BLANK" then
			normal = Indicators[prefs.NormalIndicator]
		end
	end
	if toon.Heroic and prefs.HeroicIndicator ~= "BLANK" then
		if prefs.HeroicIndicator == "TEXT" then
			heroic = ColorCodeOpen(prefs.HeroicColor) .. prefs.HeroicText .. FONTEND
		elseif prefs.HeroicIndicator ~= "BLANK" then
			heroic = Indicators[prefs.HeroicIndicator]
		end
	end
	if normal and heroic then
		return strjoin(" ", normal, heroic)
	end
	if normal then
		return normal
	end
	if heroic then
		return heroic
	end
end

local function CullStaleInstances()
	for instance, a in pairs(SavedInstances.db.Instances) do
		for toon, b in pairs(a) do
			if type(b) == "table" then
				for difficulty, c in pairs(b) do
					if c.Expires < time() then
						SavedInstances.db.Instances[instance][toon][difficulty] = nil
					end
				end
				if TableLen(b) == 0 then
					SavedInstances.db.Instances[instance][toon] = nil
				end
			end
		end
	end
end

local function HideTooltipChild()
	QTip:Release(tooltipchild)
	tooltipchild = nil
end

local function HideTooltips()
	if not tooltip or MouseIsOver(tooltip) then return end
	tooltip:SetScript("OnUpdate", nil)
	QTip:Release(tooltip)
	tooltip = nil
end

local function ShowTooltipChild(event, cell, arg)
	if not arg then return end
	if tooltipchild then QTip:Release(tooltipchild) end
	tooltipchild = QTip:Acquire("SavedInstancesTooltipChild", 1, "LEFT")
	local hFont = tooltipchild:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltipchild:SetHeaderFont(hFont)
	local instance, toon = strsplit(DELIMITER, arg)
	local thisinstance = SavedInstances.db.Instances[instance][toon]
	tooltipchild:AddHeader(strjoin("", NORMALFONT, instance, " (", FONTEND, ClassColorise(SavedInstances.db.Toons[toon].Class, strsplit(' ', toon)), NORMALFONT, ")", FONTEND))
	if thisinstance.Normal then
		tooltipchild:AddLine(strjoin("", YELLOWFONT, L["Normal ID"], ": ", FONTEND, thisinstance.Normal.ID))
		if not thisinstance.Heroic or thisinstance.Normal.Expires ~= thisinstance.Heroic.Expires then
			tooltipchild:AddLine(strjoin("", YELLOWFONT, L["Time Left"], ": ", FONTEND, SecondsToTime(thisinstance.Normal.Expires - time())))
		end
	end
	if thisinstance.Heroic then
		tooltipchild:AddLine(strjoin("", YELLOWFONT, L["Heroic ID"], ": ", FONTEND, thisinstance.Heroic.ID))
		tooltipchild:AddLine(strjoin("", YELLOWFONT, L["Time Left"], ": ", FONTEND, SecondsToTime(thisinstance.Heroic.Expires - time())))
	end
	tooltipchild:SmartAnchorTo(tooltip)
	tooltipchild:Show()
end

local function ShowTooltip(anchorframe)
	if tooltip then QTipC:Release(tooltip) end
	tooltip = QTipC:Acquire("SavedInstancesTooltip", 1, "LEFT")
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	CullStaleInstances()
	if columns then columns = nil end
	local columns = { }
	if not SavedInstances.db.Toons[thisToon].Hide and SavedInstances.db.Tooltip.FreshToons then
		columns[thisToon] = columns[thisToon] or tooltip:AddColumn("CENTER")
	end
	local headLine, headCol = tooltip:AddNormalHeader(strjoin("", NORMALFONT, "SavedInstances", FONTEND))
	for i, instance in ipairs(OrderedInstances()) do
		local currInstance = SavedInstances.db.Instances[instance]
		if TableLen(currInstance) > 2 then -- instance has toon info
			local currLine, _ = tooltip:AddNormalLine(strjoin("", GRAYFONT, instance, FONTEND))
			for toon, v in pairs(SavedInstances.db.Toons) do
				if currInstance[toon] and not SavedInstances.db.Toons[toon].Hide then
					tooltip:SetNormalCell(currLine, 1, strjoin("", NORMALFONT, instance, FONTEND))
					if not columns[toon] then
						columns[toon] = tooltip:AddColumn("CENTER")
					end	
					tooltip:SetCell(currLine, columns[toon], DifficultyString(currInstance[toon]), strjoin(DELIMITER, instance, toon))
				end
			end
		elseif SavedInstances.db.Tooltip.AllRaids and currInstance.Show then
			-- instance has no toon info, but must be shown anyway
			tooltip:AddNormalLine(strjoin("", GRAYFONT, instance, FONTEND))
		end
	end
	if tooltip:GetLineCount() == 1 then
		local noneLine = tooltip:AddLine()
		tooltip:SetNormalCell(noneLine, 1, strjoin("", GRAYFONT, NO_RAID_INSTANCES_SAVED, FONTEND), "LEFT", tooltip:GetColumnCount())
	end
	if SavedInstances.db.Tooltip.FreshToons then
		for toon, v in pairs(SavedInstances.db.Toons) do
			if not SavedInstances.db.Toons[toon].Hide then
				columns[toon] = columns[toon] or tooltip:AddColumn("CENTER")
			end
		end
	end
	for toon, col in pairs(columns) do
		tooltip:SetNormalCell(headLine, col, ClassColorise(SavedInstances.db.Toons[toon].Class, strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER")
	end 
	tooltip:AddSeparator(8,0,0,0,0)
	local hintLine, hintCol
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetNormalCell(hintLine, hintCol, L["|cffffff00Click|r to configure"], "LEFT", tooltip:GetColumnCount())
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetNormalCell(hintLine, hintCol, L["Hover mouse on indicator for details"], "LEFT", tooltip:GetColumnCount())
	tooltip:SmartAnchorTo(anchorframe)
	tooltip:Show()
	tooltip:SetCallback("OnEnter", ShowTooltipChild)
	tooltip:SetCallback("OnLeave", HideTooltipChild)
end

-- global addon code below

function SavedInstances:OnInitialize()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	if not SavedInstancesDB.DBVersion then
		SavedInstancesDB = defaultDB
	end
	if SavedInstancesDB.DBVersion == 2 then
		SavedInstancesDB.DBVersion = 3
		SavedInstancesDB.Tooltip.NewInstanceShow = false
		SavedInstancesDB.FuBar = nil
		SavedInstancesDB.profile = nil
	end
	if SavedInstancesDB.DBVersion == 3 then
		SavedInstancesDB.DBVersion = 4
		SavedInstancesDB.Indicators = defaultDB.Indicators
	end
	if SavedInstancesDB.DBVersion == 4 then
		SavedInstancesDB.DBVersion = 5
		SavedInstancesDB.Indicators.UnsavedIndicator = nil
		SavedInstancesDB.Indicators.UnsavedText = nil
		SavedInstancesDB.Indicators.UnsavedColor = nil
		for toon, a in pairs(SavedInstancesDB.Toons) do
			a.Class = strupper(gsub(a.Class, " ", ""))
			a.NumSaved = nil
		end
	end
	if SavedInstancesDB.DBVersion ~= 5 then
		SavedInstancesDB = defaultDB
	end
	self.db = db or SavedInstancesDB
	self.db.Toons[thisToon] = self.db.Toons[thisToon] or { }
	self.db.Toons[thisToon].Class = select(2, UnitClass("player"))
	self.db.Toons[thisToon].Hide = self.db.Toons[thisToon].Hide or false
	self:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(clickedframe, button)
				SavedInstances:ShowConfig()
			end,
			OnEnter = function(self)
				ShowTooltip(self)
			end,
			OnLeave = function(self)
				if tooltip and MouseIsOver(tooltip) then 
					tooltip:SetScript("OnUpdate", HideTooltips)
				else
					HideTooltips()
				end
			end,
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	CullStaleInstances()
	local numsaved = GetNumSavedInstances()
	if numsaved > 0 then
		for i = 1, numsaved do
			local iname, iid, iexpires, idiff = GetSavedInstanceInfo(i)
			if idiff > 1 then
				idiff = "Heroic"
			else
				idiff = "Normal"
			end
			self.db.Instances[iname] = self.db.Instances[iname] or { }
			self.db.Instances[iname].Order = self.db.Instances[iname].Order or TableLen(self.db.Instances)
			self.db.Instances[iname].Show = self.db.Instances[iname].Show or self.db.Tooltip.NewInstanceShow
			self.db.Instances[iname][thisToon] = self.db.Instances[iname][thisToon] or { }
			self.db.Instances[iname][thisToon][idiff] = {
				ID = iid,
				Expires = iexpires + time()
			}
		end
	end
end

-- options stuff below

SavedInstances.Options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			desc = L["General settings"],
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapIcon = {
					type = "toggle",
					name = L["Show minimap button"],
					desc = L["Show the SavedInstances minimap button"],
					order = 2,
					hidden = function() return not LDBIcon end,
					get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
					set = function(info, value)
						SavedInstances.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
					end,
				},
				FreshToons = {
					type = "toggle",
					name = L["Show non-saved characters"],
					desc = L["Show characters that have not been saved to any instances"],
					order = 3,
					get = function(info)
							return SavedInstances.db.Tooltip.FreshToons
					end,
					set = function(info, value)
							SavedInstances.db.Tooltip.FreshToons = value
					end,
				},
				List = {
					order = 4,
					type = "multiselect",
					name = L["Always hide these characters"],
					get = function(info, toon)
						return SavedInstances.db.Toons[toon].Hide
					end,
					set = function(info, toon, value)
						SavedInstances.db.Toons[toon].Hide = value
					end,
					values = function()
						local list = { }
						local k, v
						for k, v in pairs(SavedInstances.db.Toons) do
							list[k] = k
						end
						return list
					end,
				},
				Indicators = {
					order = 5,
					type = "group",
					guiInline = true,
					name = L["Status indicators"],
					desc = L["Status indicators"],
					get = function(info)
							return SavedInstances.db.Indicators[info[#info]]
					end,
					set = function(info, value)
							SavedInstances.db.Indicators[info[#info]] = value
					end,
					args = {
						NormalIndicator = {
							order = 1,
							type = "select",
							name = L["Normal indicator"],
							values = Indicators,
						},
						NormalText = {
							order = 2,
							type = "input",
							name = L["Normal text"],
							multiline = false,
							disabled = function()
								return SavedInstances.db.Indicators.NormalIndicator ~= "TEXT"
							end,
						},
						NormalColor = {
							order = 3,
							type = "color",
							hasAlpha = false,
							name = L["Normal text color"],
							disabled = function()
								return SavedInstances.db.Indicators.NormalIndicator ~= "TEXT"
							end,
							get = function(info)
								local r = SavedInstances.db.Indicators[info[#info]][1]
								local g = SavedInstances.db.Indicators[info[#info]][2]
								local b = SavedInstances.db.Indicators[info[#info]][3]
								local a = SavedInstances.db.Indicators[info[#info]][4]
								return r, g, b, a
							end,
							set = function(info, r, g, b, a)
								SavedInstances.db.Indicators[info[#info]][1] = r
								SavedInstances.db.Indicators[info[#info]][2] = g
								SavedInstances.db.Indicators[info[#info]][3] = b
								SavedInstances.db.Indicators[info[#info]][4] = a
							end
						},
						HeroicIndicator = {
							order = 4,
							type = "select",
							name = L["Heroic indicator"],
							values = Indicators,
						},
						HeroicText = {
							order = 5,
							type = "input",
							name = L["Heroic text"],
							multiline = false,
							disabled = function()
								return SavedInstances.db.Indicators.HeroicIndicator ~= "TEXT"
							end,
						},
						HeroicColor = {
							order = 6,
							type = "color",
							hasAlpha = false,
							name = L["Heroic text color"],
							disabled = function()
								return SavedInstances.db.Indicators.HeroicIndicator ~= "TEXT"
							end,
							get = function(info)
								local r = SavedInstances.db.Indicators[info[#info]][1]
								local g = SavedInstances.db.Indicators[info[#info]][2]
								local b = SavedInstances.db.Indicators[info[#info]][3]
								local a = SavedInstances.db.Indicators[info[#info]][4]
								return r, g, b, a
							end,
							set = function(info, r, g, b, a)
								SavedInstances.db.Indicators[info[#info]][1] = r
								SavedInstances.db.Indicators[info[#info]][2] = g
								SavedInstances.db.Indicators[info[#info]][3] = b
								SavedInstances.db.Indicators[info[#info]][4] = a
							end
						},
					},
				},
			},
		},
		Instances = {
			order = 4,
			type = "group",
			name = L["Instance options"],
			desc = L["Instance options"],
			args = {
				OrderGroup = {
					order = 1,
					type = "group",
					guiInline = true,
					name = L["Instance order"],
					desc = L["Instance order"],
					get = function(info, instance)
						return SavedInstances.selectedInstance
					end,
					args = {
						Selected = {
							order = 1,
							type = "select",
							name = L["Instance to be moved"],
							set = function(info, instance)
								SavedInstances.selectedInstance = instance
							end,
							values = OrderedInstances,
						},
						Position = {
							order = 2,
							type = "select",
							name = L["Position in list"],
							set = function(info, instance)
								SetInstanceOrder(SavedInstances.selectedInstance, instance)
								SavedInstances.selectedInstance = nil
							end,
							values = OrderedInstances,
						},
					},
				},
				VisibilityGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Instance visibility"],
					desc = L["Instance visibility"],
					get = function(info)
							return SavedInstances.db.Tooltip[info[#info]]
					end,
					set = function(info, value)
							SavedInstances.db.Tooltip[info[#info]] = value
					end,
					args = {
						AllRaids = {
							type = "toggle",
							order = 1,
							name = L["Show all instances"],
							desc = L["Always show all instances, even if no characters are saved to them"],
						},
						NewInstanceShow = {
							type = "toggle",
							order = 2,
							name = L["Always show new instances"],
							desc = L["New instances are shown by default, even if no characters are saved to them"],
							disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
						},
						List = {
							type = "multiselect",
							order = 3,
							name = L["Show these instances when all instances are shown"],
							disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
							get = function(info, instance)
								local trimmedInstance = strsub(instance, 5)
								return SavedInstances.db.Instances[trimmedInstance].Show
							end,
							set = function(info, instance, value)
								local trimmedInstance = strsub(instance, 5)
								SavedInstances.db.Instances[trimmedInstance].Show = value
							end,
							values = function()
								local list = { }
								local i, v
								for i, v in ipairs(OrderedInstances()) do
									local prefix = format("%04d", i)
									list[prefix .. v] = v
								end
								return list
							end,
						},
					},
				},
			},
		},
	},
}

local optionsFrames = { }
function SavedInstances:SetupOptions()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", SavedInstances.Options)
	optionsFrames["General"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", nil, nil, "General")
	optionsFrames["Instances"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Instance options"], "SavedInstances", "Instances")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["Instances"])
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["General"])
end

